import torch
import torch.nn as nn
import torch.nn.functional as F

"""
Default architecture use for the formations.
"""


class ClassicNet(nn.Module):
    def __init__(self, num_layers):
        super().__init__()
        layers = [nn.Conv2d(3, 100, 3, padding=1), nn.ReLU()]
        self.num_layers = num_layers
        for layer_number in range(2, num_layers + 1):
            layers.append(nn.Conv2d(100, 100, 3, padding=1))
            layers.append(nn.ReLU())
            if layer_number in [3, 5]:
                layers.append(nn.MaxPool2d(2))

        layers.append(nn.Flatten())
        layers.append(nn.Linear(100 * 8 * 8, 10))

        self.model = nn.Sequential(*layers)

    def forward(self, x):

        for layer in self.model:
            x = layer(x)

        return x

    def eval_weight_distribution(self, x):
        k = 2000  # Chiffre de valeurs aléatoires
        entree = dict()
        indice = 0
        for layer in self.model:
            if isinstance(layer, torch.nn.modules.conv.Conv2d):
                indice += 1
                if (
                        indice == 1 or indice == self.num_layers - 1
                ):  # Première ou dernière couche

                    echantillon = nn.Flatten(0)(x.detach().cpu())
                    entree[f"Conv_{indice}"] = np.random.choice(echantillon, size=k)

            x = layer(x)

        return x, entree


class ResidualLayer(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(100, 100, 3, padding=1)

        self.conv2 = nn.Conv2d(100, 100, 3, padding=1)

    def forward(self, x):
        res = x
        x = F.relu((self.conv1(x)))
        x = self.conv2(x)
        x = F.relu(x + res)
        return x


class ResNet(nn.Module):
    def __init__(self, nb_bloc_residuelle):
        super().__init__()
        assert nb_bloc_residuelle > 2

        self.conv1 = nn.Conv2d(3, 100, 3, padding=1)
        self.bn1 = nn.BatchNorm2d(100)

        self.layers = nn.ModuleList(
            [ResidualLayer() for _ in range(nb_bloc_residuelle)]
        )

        self.fc1 = nn.Linear(100 * 8 * 8, 10)

    def forward(self, x):
        x = F.relu(self.bn1(self.conv1(x)))

        for i, layer in enumerate(self.layers):
            x = layer(x)
            if i <= 1:
                x = F.max_pool2d(x, 2)

        x = x.flatten(1)
        x = self.fc1(x)
        return x



import time

from sklearn.metrics import accuracy_score


def evaluate(pipeline, train, test):
    """
    Cette fonction sera utilisée pour comparer la performance,
    le temps d'entraînement et la taille du vocabulaire des différents pipelines
    """
    train_data, train_label = train
    test_data, test_label = test

    start_time = time.time()
    pipeline.fit(train_data, train_label)
    end_time = time.time()

    predict_train = pipeline.predict(train_data)
    predict_test = pipeline.predict(test_data)

    accuracy_train = accuracy_score(predict_train, train_label)
    accuracy_test = accuracy_score(predict_test, test_label)
    train_time = end_time - start_time

    return accuracy_train, accuracy_test, train_time


def evaluate_20newsgroups(pipeline, train, test):
    """
    Une fonction pour évaluer sur 20newsgroup
    """
    return evaluate(pipeline, (train.data, train.target), (test.data, test.target),)


def evaluate_weccr(pipeline, train, test):
    """
    Une fonction pour évaluer sur WECCR
    """
    return evaluate(
        pipeline, (train["text"], train["label"]), (test["text"], test["label"])
    )



import numpy as np
import torch
import matplotlib.pyplot as plt
import math
from torch.utils.data import Dataset
import pandas as pd
import pickle as pk


class classification_jeu_donnee:
    def __init__(self, taille_echantillon=100, bruit_bleu=0.3, bruit_rouge=0.5):

        (
            self.data,
            self.x_bleu,
            self.y_bleu,
            self.x_rouge,
            self.y_rouge,
        ) = self._generate_dataset(taille_echantillon, bruit_bleu, bruit_rouge)

    def _generate_dataset(self, taille_echantillon, bruit_bleu, bruit_rouge):
        x = np.random.rand(taille_echantillon) * 10 - 5
        y = np.random.rand(taille_echantillon) * 30
        borne_linaire = np.linspace(-2.3166, 3, 100)
        borne_quadratique_positive = np.linspace(-(12.5 ** 0.5), -2.3166, 100)
        borne_quadratique_negative = np.linspace(-(12.5 ** 0.5), 3, 100)
        x_bleu = list()
        y_bleu = list()
        x_rouge = list()
        y_rouge = list()

        for i, j in zip(x, y):
            if (
                    (j > i ** 2) and (j < -(i ** 2) + 25) and (j > i * 2 + 10)
            ):  # or random.random() <= 0.02) :
                x_rouge = np.append(x_rouge, i)
                y_rouge = np.append(y_rouge, j)
            else:

                x_bleu = np.append(x_bleu, i)
                y_bleu = np.append(y_bleu, j)

        x_bleu = x_bleu + (np.random.randn(len(x_bleu)) * bruit_bleu)
        y_bleu = y_bleu + (np.random.randn(len(y_bleu)) * bruit_bleu)
        x_rouge = x_rouge + (np.random.randn(len(x_rouge)) * bruit_rouge)
        y_rouge = y_rouge + (np.random.randn(len(y_rouge)) * bruit_rouge)
        data_classification_plan = list()
        for i, j in zip(x_bleu, y_bleu):
            data_classification_plan.append(
                (
                    torch.Tensor(
                        [i, j, i ** 2, j ** 2, i * j, math.sin(i), math.sin(j)]
                    ),
                    0,
                )
            )

        for i, j in zip(x_rouge, y_rouge):
            data_classification_plan.append(
                (
                    torch.Tensor(
                        [i, j, i ** 2, j ** 2, i * j, math.sin(i), math.sin(j)]
                    ),
                    1,
                )
            )

        return data_classification_plan, x_bleu, y_bleu, x_rouge, y_rouge

    def visualisation(self):
        borne_linaire = np.linspace(-2.3166, 3, 100)
        borne_quadratique_positive = np.linspace(-(12.5 ** 0.5), -2.3166, 100)
        borne_quadratique_negative = np.linspace(-(12.5 ** 0.5), 3, 100)
        plt.scatter(self.x_bleu, self.y_bleu, s=3)
        plt.scatter(self.x_rouge, self.y_rouge, marker="D", s=3)

        plt.plot(borne_quadratique_positive, borne_quadratique_positive ** 2, c="r")
        plt.plot(
            borne_quadratique_negative, -(borne_quadratique_negative ** 2) + 25, c="r"
        )
        plt.plot(borne_linaire, 2 * borne_linaire + 10, c="r")
        plt.xlim(-5, 5)
        plt.ylim(0, 30)
        plt.show()

    def __getitem__(self, idx):
        return self.data_classification_plan[idx]

    def __len__(self):
        return len(self.data_classification_plan)


class FaceDataset(Dataset):
    def __init__(self, csv_file, transform=None, columns=None):

        self.data = pd.read_csv(csv_file)

        if columns is None:
            self.columns = ["age", "ethnicity", "gender"]
        else:
            self.columns = columns

        self.data.drop(columns={"img_name"}, inplace=True)
        self.data["pixels"] = self.data["pixels"].apply(
            lambda x: np.array(x.split(), dtype="float32").reshape((1, 48, 48)) / 255
        )
        self.data["age"] = self.data["age"].apply(
            lambda x: np.array([x], dtype="float32")
        )
        self.X = torch.Tensor(self.data["pixels"])

        self.transform = transform

    def __len__(self):
        return len(self.data)

    def __getitem__(self, indice):
        if torch.is_tensor(indice):
            indice = indice.tolist()
        image = self.X[indice]
        if len(self.columns) > 1:
            attribute = torch.Tensor(
                [float(self.data.iloc[indice][i]) for i in self.columns]
            )
        else:
            attribute = self.data.iloc[indice][self.columns[0]]
        sample = (image, attribute)

        if self.transform:
            sample = (self.transform(sample[0]), attribute)

        return sample


class EchantillonCIFAR10(Dataset):
    """
        Échantillon du jeu de donnée CIFAR10. L'échantillon comprend 10 000 données d'entraînements
        et 2 000 de tests.
    Args:
        root_dir (string): Chemin vers le fichier pickle contenant l'échantillon
        train (bool): Si True, prend les données d'entraînements, sinon les données de tests
        transforms (callable, optional): Une function/transform qui prend le target et le transforme.

    """

    def __init__(self, root_dir, train=True, transform=None):

        if train:
            self.echantillon = pk.load(
                open(f"{root_dir}CIFAR10_train_10000_sample.pk", "rb")
            )
        else:
            self.echantillon = pk.load(
                open(f"{root_dir}CIFAR10_test_2000_sample.pk", "rb")
            )

        self.root_dir = root_dir
        self.transform = transform
        self.classes = [
            "avion",
            "automobile",
            "oiseau",
            "chat",
            "chevreuil",
            "chien",
            "grenouille",
            "cheval",
            "bateau",
            "camion",
        ]

    def __len__(self):
        return len(self.echantillon)

    def __getitem__(self, indice):

        if torch.is_tensor(indice):
            indice = indice.tolist()
        img, target = self.echantillon[indice]
        if self.transform is not None:
            img = self.transform(img)

        return img, target



import os
import subprocess

from setuptools import setup, find_packages

current_file_path = os.path.abspath(os.path.dirname(__file__))


def get_readme():
    readme_file_path = os.path.join(current_file_path, "README.md")
    with open(readme_file_path, "r", encoding="utf-8") as f:
        return f.read()


def get_version():
    version_file_path = os.path.join(current_file_path, "version.txt")
    with open(version_file_path, "r", encoding="utf-8") as f:
        version = f.read().strip()

    try:
        sha = subprocess.check_output(["git", "rev-parse", "HEAD"]).decode("ascii").strip()
    except Exception:  # pylint: disable=broad-except
        sha = "Unknown"

    if os.getenv("PAX_LIBRARIES_RELEASE_BUILD") != "1":
        version += ".dev1"
        if sha != "Unknown":
            version += "+" + sha[:7]
    return version


def write_version_python_file(version):
    version_python_file = os.path.join(current_file_path, "paxLibUL", "version.py")
    with open(version_python_file, "w", encoding="utf-8") as f:
        f.write(f"__version__ = {repr(version)}\n")


def main():
    readme = get_readme()

    version = get_version()
    print("Building version", version)
    write_version_python_file(version)

    packages = find_packages()
    setup(
        name="paxLibUL",
        version=version,
        author="PaX-UL",
        url="https://pax.ulaval.ca/",
        download_url="https://github.com/PAX-ULaval/pax-libraries/archive/v" + version + ".zip",
        license="LGPLv3",
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Topic :: Software Development :: Libraries",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
        packages=packages,
        install_requires=[
            "sklearn",
            "torch",
            "poutyne",
            "numpy",
            "matplotlib",
            "pandas",
            "seaborn",
        ],
        python_requires=">=3.7",
        description="A library for internal formation tools use in PaX ULaval formations.",
        long_description=readme,
        long_description_content_type="text/markdown",
    )


if __name__ == "__main__":
    main()



from poutyne import Callback


class sauvegardeLR(Callback):
    def __init__(self, optimizer):
        super().__init__()
        self.optimizer = optimizer
        self.lr_par_epoch = list()

    def on_epoch_begin(self, batch, logs):  ## À chaque début d'époque
        self.lr_par_epoch.append(self.optimizer.state_dict()["param_groups"][0]["lr"])



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch

from poutyne import Callback


def visualiser_difference_de_poids(poids_initiaux, poids_finaux):
    """
    Affiche un graphique de la différence absolue moyenne entre les poids initiaux et finaux par couche.
    Args:
        poids_initiaux (Dict[str, torch.Tensor] ): Un dictionnaire où les clés sont le nom de la couche et leur valeur sont les poids
        poids_finaux (Dict[str, torch.Tensor] ): Un dictionnaire où les clés sont le nom de la couche et leur valeur sont les poids
    """

    difference_etat = dict()
    nb_couches = len(poids_initiaux.keys())

    for etat_initial_couche, etat_final_couche in zip(
        poids_initiaux.items(), poids_finaux.items()
    ):
        difference_etat[etat_initial_couche[0]] = torch.abs(
            torch.flatten(
                (etat_final_couche[1].cpu() - etat_initial_couche[1].cpu())
                .detach()
                .clone(),
                0,
            )
        ).tolist()

    moyenne = [np.mean(poids_couche) for poids_couche in difference_etat.values()]
    couche = [np.mean(i) for i in range(len(difference_etat.values()))]

    max_moyenne = np.max(moyenne)

    sns.lineplot(couche, moyenne)  # ,ax=ax2)
    plt.title(
        f"Valeur moyenne de la différence absolue des poids initiaux et finaux par couche dans une architecture de {nb_couches} couches"
    )
    plt.xlabel("Indice de la couche")
    plt.ylabel("différence")
    plt.ylim(0, max_moyenne)
    plt.show()


class SauvegarderPoids(Callback):
    """
    This callback multiply the loss temperature with a decay before
    each batch.

    Args:
        celoss_with_temp (CrossEntropyLossWithTemperature): the loss module.
        decay (float): The value of the temperature decay.
    """

    def __init__(self):
        super().__init__()
        self.liste_poids = list()

    def on_epoch_begin(self, batch, logs):  ## À chaque début d'époque
        poids = tuple(self.model.network.parameters())[0][0].clone().detach().cpu()
        self.liste_poids.append(poids)


def show_2d_function(
    fct,
    min_val=-5,
    max_val=5,
    mesh_step=0.01,
    *,
    optimal=None,
    bar=True,
    ax=None,
    **kwargs,
):
    # pylint: disable=blacklisted-name
    """
    Trace les courbes de niveau d'une fonction 2D.
    Args:
        fct (Callable[torch.Tensor, torch.Tensor]): Fonction objectif qui prend en paramètre un tenseur Nx2
            correspondant à N paramètres pour lesquels on veut obtenir la valeur de la fonction.
        optimal (torch.Tensor): La valeur optimale des poids pour la fonction objectif.
    """
    w1_values = torch.arange(min_val, max_val + mesh_step, mesh_step)
    w2_values = torch.arange(min_val, max_val + mesh_step, mesh_step)

    w2, w1 = torch.meshgrid(w2_values, w1_values)
    w_grid = torch.stack((w1.flatten(), w2.flatten()))
    fct_values = fct(w_grid).view(w1_values.shape[0], w2.shape[0]).numpy()

    w1_values, w2_values = w1_values.numpy(), w2_values.numpy()

    if ax is not None:
        plt.sca(ax)
    if "cmap" not in kwargs:
        kwargs["cmap"] = "RdBu"
    plt.contour(w1_values, w2_values, fct_values, 40, **kwargs)
    plt.xlim((min_val, max_val))
    plt.ylim((min_val, max_val))
    plt.xlabel("$w_1$")
    plt.ylabel("$w_1$")

    if bar:
        plt.colorbar()

    if optimal is not None:
        plt.scatter(*optimal.numpy(), s=200, marker="*", c="r")


def show_2d_trajectory(
    w_history, fct, min_val=-5, max_val=5, mesh_step=0.5, *, optimal=None, ax=None
):
    """
    Trace le graphique de la trajectoire de descente en gradient en 2D.
    Args:
        w_history: L'historique de la valeur des poids lors de l'entraînement.
        fct (Callable[torch.Tensor, torch.Tensor]): Fonction objectif qui prend en paramètre un tenseur Nx2
            correspondant à N paramètres pour lesquels on veut obtenir la valeur de la fonction.
        optimal (torch.Tensor): La valeur optimale des poids pour la fonction objectif.
    """
    show_2d_function(fct, min_val, max_val, mesh_step, optimal=optimal, ax=ax)

    if len(w_history) > 0:
        trajectory = np.array(w_history)
        plt.plot(trajectory[:, 0], trajectory[:, 1], "o--", c="g")

    plt.title("Trajectoire de la descente en gradient")
    plt.xlabel("$w_1$")
    plt.ylabel("$w_2$")


def show_learning_curve(loss_list, loss_opt=None, ax=None):
    """
    Trace le graphique des valeurs de la fonction objectif lors de l'apprentissage.
    Args:
        loss_list: L'historique de la valeur de la perte lors de l'entraînement.
        loss_opt: La valeur optimale de perte.
    """
    if ax is not None:
        plt.sca(ax)
    plt.plot(
        np.arange(1, len(loss_list) + 1),
        loss_list,
        "o--",
        c="g",
        label="$F(\\mathbf{w})$",
    )
    if loss_opt is not None:
        plt.plot([1, len(loss_list)], 2 * [loss_opt], "*--", c="r", label="optimal")
    plt.title("Valeurs de la fonction objectif")
    plt.xlabel("Itérations")
    plt.legend()


def show_optimization(w_history, loss_history, fct, optimal=None, title=None):
    """
    Trace deux graphiques montrant le trajet de l'optimisation d'une fonction objectif 2D. Le premier montre la valeur
    des poids lors de l'optimisation. Le deuxième montre la valeur de la perte lors de l'optimisation.
    Args:
        w_history: L'historique des poids lors de l'optimisation
        loss_history: L'historique de la valeur de la fonction perte.
        fct (Callable[torch.Tensor, torch.Tensor]): Fonction objectif qui prend en paramètre un tenseur Nx2
            correspondant à N paramètres pour lesquels on veut obtenir la valeur de la fonction.
        optimal (torch.Tensor): La valeur optimale des poids pour la fonction objectif.
    """
    fig, axes = plt.subplots(1, 2, figsize=(14.5, 4))
    if title is not None:
        fig.suptitle(title)
    show_2d_trajectory(w_history, fct, optimal=optimal, ax=axes[0])
    show_learning_curve(loss_history, loss_opt=fct(optimal).numpy(), ax=axes[1])


def plot_image_samples(dataset):
    """
    Affiche certaines images d'un jeu de données
    """
    classes = dataset.classes
    num_samples = 15
    indices = np.random.choice(len(dataset), size=num_samples, replace=False)
    NROWS = 3
    NCOLS = 5

    assert (
        len(indices) <= NROWS * NCOLS
    ), f"Impossible d'afficher plus de {NROWS * NCOLS} images."

    samples = [dataset[i] for i in indices]
    images, targets = zip(*samples)
    targets = list(targets)

    # Affichage des images avec leurs étiquettes
    fig, axes = plt.subplots(NROWS, NCOLS, sharex=True, sharey=True)
    for ax, i, image, target in zip(axes.flat, indices, images, targets):
        ax.imshow(image, cmap="gray")

        target_class = classes[target]
        ax.set_xlabel(f"Exemple {i}:\n" + target_class, fontsize=8)
        ax.set_xticks([])
        ax.set_yticks([])

    fig.tight_layout()
    plt.show()


def visualiser_convolution(reseau, donnee, nb_image_max_par_couche):
    x = torch.unsqueeze(list(donnee)[0][0], 0)
    image_initiale = x[0].numpy().transpose((1, 2, 0))
    x = x.cuda()
    print("Image initiale :")
    plt.imshow(image_initiale)
    plt.show()
    with torch.no_grad():
        conv = 0
        pool = 0

        for layer in reseau.children():
            x = layer(x)

            if isinstance(layer, torch.nn.Conv2d):
                conv += 1
                print(f"Convolution {conv}")
                _, axes = plt.subplots(
                    min(x.shape[1] // 4, nb_image_max_par_couche // 4),
                    4,
                    constrained_layout=True,
                )
                for feature in range(min(x.shape[1], nb_image_max_par_couche)):
                    image_conv = x[0, feature, :, :].clone().cpu()

                    try:
                        axes[feature // 4, feature % 4].imshow(image_conv)
                    except IndexError:
                        axes[feature % 4].imshow(image_conv)
                plt.show()
            if isinstance(layer, torch.nn.modules.pooling.MaxPool2d) or isinstance(
                layer, torch.nn.modules.pooling.AvgPool2d
            ):
                pool += 1
                print(f"Pooling {pool}")
                _, axes = plt.subplots(
                    min(x.shape[1] // 4, nb_image_max_par_couche // 4),
                    4,
                    constrained_layout=True,
                )
                for feature in range(min(x.shape[1], nb_image_max_par_couche)):
                    imageConv = x[0, feature, :, :]
                    try:
                        axes[feature // 4, feature % 4].imshow(image_conv)
                    except IndexError:
                        axes[feature % 4].imshow(image_conv)

                plt.show()
